!------------------------------------------------------------------------------------------------------------------
!  This file contains all the necessary functions and subroutines grouped into modules
!  Almost all the following functions are simply the numerical equivalents of the terms discussed in the lab manual
!
!  Your work has to be done in the module chiscan and chain which are both located at the end of this file
!------------------------------------------------------------------------------------------------------------------

!###########################################################################################################################
! Here we simply define some physical constants
module constants
  implicit none

  ! The constants
  double precision, parameter :: ckms = 3.0d5                      ! speed of light in km/s
  double precision, parameter :: rhocrit0 = 2.7751973751261264d11  ! rhocrit in units of h^-1 Msun/ h^-3 Mpc^3
  double precision, parameter :: Gs = 4.3d-9                       ! G in units of Mpc/Msun /(km/s)^2
  double precision, parameter :: pi = 4.0d0 * atan(1.0d0)

  ! Once the redshift of the lens is known these numbers are constants throughtout the entire code
  ! hinted by the keyword 'SAVE'
  double precision, save :: dAlens,rhoclens

end module constants
!###########################################################################################################################



!###########################################################################################################################
! Definition of the cosmological model
! It contains actually a lot of different routines 
module cosmology
  use constants
  implicit none

  ! The cosmological parameters
  ! For a flat universe the length in units of (Mpc * h^-1)is straigtforward
  public
  type cospar
     double precision :: omegam
  end type cospar
  type(cospar), save :: cosmopar
  
  
  contains
  ! Calculate the Hubble parameter 
  ! E=H(z)/H0
  function Eh(z)
    ! Defining the variables
    double precision :: Eh, a
    double precision, intent(in) :: z  ! The value of 'z' is input
    
    a = 1.0d0 / (1.0d0+z)
    Eh = dsqrt((cosmopar%omegam / (a*a*a)) + (1.0d0-cosmopar%omegam))
    
    return
  end function Eh


  ! A USEFUL ROMBERG INTEGRATION ROUTINE
  function rombint(f,a,b,tol)
  ! Rombint returns the integral from a to b of using Romberg integration.
  ! The method converges provided that f(x) is continuous in (a,b).
  ! f must be DOUBLE PRECISION and must be declared external in the calling routine.  
  ! tol indicates the desired relative accuracy in the integral.

  ! The routine won't be explained in detail here, but the two main points are:
  ! 1: Use the trapeziodal rule to get a first (rough) approximation
  ! 2: Use the 'Richardson-Romberg' method to step-by-step increase the accuracy of the calculation
  implicit none
    
    ! Defining the variables
    integer, parameter :: maxiter = 20, maxj = 5
    dimension g(MAXJ+1)
    double precision, intent(in) :: a, b, tol
    double precision :: f, rombint, h, gmax, error, g, g0, g1, fourj
    external f
    integer :: nint, i, k, jmax, j
    
    ! Initialize the integration routine
    h = 0.5d0 * (b-a)
    gmax = h * (f(a) + f(b))
    g(1) = gmax
    nint = 1
    error = 1.0d20
        i = 0
10        i = i+1
          if (i .gt. maxiter .or. (i .gt. 5 .and. abs(error) .lt. tol)) &
            go to 40


    ! Calculate next trapezoidal rule approximation to integral
     g0 = 0.0

     do 20 k = 1, nint
       g0 = g0 + f(a + (k+k-1) * h)
20     continue
       g0 = 0.5d0*g(1) + h*g0
       h = 0.5d0 * h
       nint = nint + nint
       jmax = min(i,maxj)
       fourj = 1.0
         
       do 30 j = 1, jmax
         ! Use Richardson extrapolation.
         fourj = 4.0 * fourj
         g1 = g0 + (g0-g(j)) / (fourj-1.0)
         g(j) = g0
         g0 = g1
30       continue
         if (abs(g0) .gt. tol) then
           error = 1.0 - gmax/g0
         else
           error = gmax
         end if
         gmax = g0
         g(jmax+1) = g0
         go to 10
40       rombint = g0
         if (i .gt. maxiter .and. abs(error). gt.tol)  then
           write(*,*) 'Warning: Rombint failed to converge; '
           write(*,*)'integral, error, tol:', rombint,error, tol
        endif
        
  end function rombint


  !-----------------------------------------------------------------------
  ! Calculate the coordinate distance , in units of h^-1 Mpc
  ! Use the Romberg integration method to do this
  function r(z)
    ! Defining the variables
    double precision, intent(in) :: z
    double precision, parameter :: tol = 1.0d-8
    double precision :: r, integral

    integral = rombint(rint,0.0d0,z,tol)
    r = (ckms*integral) / 1.0d2
    
    return
  end function r


  !-----------------------------------------------------------------------
  ! This function simply defines integrand for coordinate distance 
  function rint(z)
    double precision, intent(in) :: z
    double precision :: rint

   rint = 1.0d0 / Eh(z)
   
   return
  end function rint


  !-----------------------------------------------------------------------
  ! Here we calcualte the angular diameter distance in units of h^-1 Mpc
  function dA(z)
    ! Defining the varuables 
    double precision, intent(in) :: z
    double precision :: dA
    
    dA = r(z) / (1.+z)
    
    return
  end function dA


  !-----------------------------------------------------------------------    
  ! Calculate the  critical density as a function of redshift
  function rhocrit(z)
    ! Defining the variables
    double precision, intent(in) :: z
    double precision :: rhocrit, hdim
    
    hdim = Eh(z)
    rhocrit = rhocrit0 * hdim * hdim

    return
  end function rhocrit
end module cosmology
!###########################################################################################################################



!###########################################################################################################################
! This module calculate some quantities of the density profile for the lenssystem 
module profile
  use constants
  use cosmology
  implicit none
  
  contains
  ! Critical Surface Density in units of hMsun/Mpc^2
  function Sigmac(zd,zs)
    double precision, intent(in) :: zd, zs
    double precision :: Sigmac, Ds, Dd, Dds

    Ds = dA(zs)
    Dd = dAlens

    ! The following only works for flat Universe
    Dds = Ds - (1.0d0+zd) * Dd/(1.0d0+zs)
    Sigmac = (0.25d0*ckms*ckms*s) / (Dd*Dds*pi*Gs)
      
    return
  end function Sigmac


  !-------------------------------------------------------------------------        
  ! Here we calculate the density contrast deltac of the NFW density profile
  function deltac(conc)

    double precision, intent(in) :: conc
    double precision :: deltac
   
    deltac = (200.0d0*conc*conc*conc) / ((dlog(1.0+conc) - conc/(1.0d0+conc)) * 3.0d0)

    return
  end function deltac


  !-----------------------------------------------------------------------    
  ! Calculation of the 'g-part' of  the shear calculation
  function ggam(x)
    double precision, intent(in) :: x  
    double precision :: ggam, ath, arct
    
    if (x < 1.0d0) then
      ath = datanh(dsqrt((1.0d0-x) / (1.0d0+x)))
      ggam = 8.0d0 * ath/((x*x) * dsqrt(1.0d0-x*x)) + 4.0d0/(x*x)*dlog(x/2.0d0) - 2.0d0 / (x*x-1.0d0) + &
             4.0d0 * ath/((x*x-1.0d0) * dsqrt(1.0d0-x*x))
    elseif (x > 1.0d0) then
      arct = datan(dsqrt((x-1.0d0)/(1.0d0+x)))
      ggam = 8.0d0 * arct/((x*x) * dsqrt(x*x-1.0d0)) + 4.0d0/(x*x)*dlog(x/2.0d0) - 2.0d0/(x*x-1.0d0) + &
             4.0d0 * arct/(x*x-1.0d0)**1.5d0
    else
      ggam = 10.0d0/3.0d0 - 4.0d0*dlog(2.0d0) ! The formula in the manual actual goes like: '+ 4*ln(0.5)'
    endif

    return
  end function ggam


  !----------------------------------------------------------------------------------------------------------    
  ! Here we calculate the tangential shearing
  ! You will need this function to calculate the chi^2 later on, the parameters of gshear are explained there
  function gshear(x,zd,zs,rs,dc)
    double precision, intent(in) :: x, zd, zs, rs, dc  
    double precision :: gshear
            
    gshear = rs*dc*rhoclens*ggam(x) / Sigmac(zd,zs)
  
    return
  end function gshear
end module profile
!###########################################################################################################################



!###########################################################################################################################
! This module manages the input of files and parameters
module inputdata
  use cosmology
  use constants
  use profile
  implicitn one
  
  double precision, allocatable, save :: zdat(:), sheardat(:), thetadat(:), rdat(:)
  double precision, save :: sigmashear, zlens
  double precision, save :: zlens

  integer, save :: ndat

  contains
  !-----------------------------------------------------------------------    
  ! This subroutine reads in (the) inputfile(s)
  subroutine readinput(filename)
      character(80) :: filename
      integer :: i, idummy
      double precision :: ddummy
      character(80) :: dummystring
      
      ! Open the output file - The name comes from the 'main' program
      open(unit=12, file=filename, status='unknown', form='formatted')
      read(12,*) dummystring
      READ(12,*) dummystring
      
      ! Should any arrays of the following arrays be allocated, we clear them up here
      if (allocated(zdat))     deallocate(zdat)
      if (allocated(sheardat)) deallocate(sheardat)
      if (allocated(thetadat)) deallocate(thetadat)
      if (allocated(rdat))     deallocate(rdat)
      
      ! Assign the arrays with the dimension given by the value of 'ndat'
      allocate(thetadat(ndat),zdat(ndat),sheardat(ndat),rdat(ndat))
      
      ! Read in the file
      do i=1, ndat
        read(12,*) idummy,ddummy,ddummy,thetadat(i),ddummy,sheardat(i), & 
                   ddummy,ddummy,ddummy,ddummy,ddummy,ddummy,ddummy,zdat(i)
      enddo

      ! Close the file
      close(12)

      !!dir$ parallel
      !!dir$ loop count min(256)
      rdat = (dA(zlens)*thetadat*pi) / (180.0d0*60.0d0)

      return
      
    end subroutine readinput

!-----------------------------------------------------------------------    
! This function calculates the scale radius for given mass (M_200 w.r.t. critical density) and concentration
    function rsnfw(M200,con)
      double precision, intent(in) :: M200,con
      double precision :: rsnfw
      
      rsnfw = (3.0d0*M200 / (800.0d0*pi*rhoclens*con*con*con))**(1.0d0/3.0d0)

    end function rsnfw
end module inputdata
!###########################################################################################################################



!###########################################################################################################################
! This module is the important one for the first exercise, as here the chi^2 is calculated
! BEWARE: some parts of the input come as arrays, while others are simply numbers 
module chi2calc
  use cosmology
  use inputdata
  implicit none

  contains
  !-----------------------------------------------------------------------    
  function chi2(rs,dc)
    double precision, intent(in) :: rs, dc  ! rs is the scale radius | dc is the critical density
    double precision :: chi2, x, zd, zs     ! x is the dimensionless radial distance | zd holds the redshift of the lense
                                            ! and zs stores the redshifts of the galaxies
    integer :: i

    ! Here comes your implementation of chi^2
    ! You should use the function gshear and the sheardat array
    chi2 = 0.
    
    ! PLEASE IMPLEMENT ME
    
    return
  end function chi2
end module chi2calc
!###########################################################################################################################




!###########################################################################################################################
! This module contains the Monte Carlo Markov chain algorithm
! It is the second big implementation part of the lab
module mcmc
  integer, save :: npar, nsample
  double precision, allocatable :: par(:), dpar(:), lobound(:), upbound(:), dpari(:)  ! The arrays aren't initialized here

  contains
  !--------------------------------------------------------
  ! This first subroutine generates the Gaussian RNG, using the instrinsic Fortran routine 'random number'
  subroutine gaussrandom(a)
  SUBROUTINE GAUSSRANDOM(A)
    double precision :: a(i)
    double precision :: v1, v2, rsq, fac
    real :: r1, r2
    integer :: ndim, i
    
    ndim = size(a)
    NDIM=SIZE(A)
    
    do i=1, ndim
      rsq = 1.5
      ! Here we generate the random numbers
      do while ((rse .gt. 1.0) .or. (rsq .eq. 0.0))
        call random_number(r1)
        call random_number(r2)
        v1 = 2.0 * dble(r1) - 1.0d0
        v2 = 2.0 * dble(r2) - 1.0d0
        rsq = (v1*v1) + (v2*v2)
      enddo
      
      fac = dsqrt((-2.0d0 * dlog(rsq)) / rsq)
      a(i) = v1 * fac
    enddo  
  end subroutine gaussrandom

  !------------------------------------------------------------------------------------------------
  ! The following is the first part of the actual chain routine, which initializes the whole thing

  ! Initialize random_seed based on system time
  subroutine init_random_seed() 
    integer :: i, n, clock
    integer, dimension(:), allocatable :: seed
    
    ! random_seed is also an intrinsic routine
    ! initialized correctly it will give a different random number sequence with each invocation
    call random_seed(size=n)
    ! Assign the memory space to the seed array
    allocate(seed(n))
    
    ! system_clock gives the time difference between now and somewhere in the past
    call system_clock(count=clock)

    seed = clock + 37 * (/(i-1, i=1, n) /)
    
    call random_seed(put=seed)

    ! Free the memory space
    deallocate(seed)
  end subroutine init_random_seed
  
  !----------------------------------------------------------------------------------------------
  ! This subroutine contains the part of the MCMC you have to implement 
  ! by following the 'pseudo-code' given in the lab manual
  subroutine chain(filename)
    integer :: i, np, count
    double precision :: chi2get, chiold, accept, atest
    double precision, allocatable :: parold(i), rshift(:)
    character(80) :: filename
    
    ! Assign the memory space to the arrays
    allocate (parold(npar),rshift(npar))
    
    ! Beginning of the MCMC algorithm
    ! Initialize the RNG
    !--------------------------------------------------------------------------!
    ! IN THE FOLLOWING, MAKE USE OF THE TWO SUBROUTINES FOLLOWING THIS ROUTINE !
    !--------------------------------------------------------------------------!
    CALL init_random_seed()
    
    ! HEY MAN, I WANT TO BE CODED
    !......

    deallocate(parold,rshift)
    
    return
  end subroutine chain  

  !---------------------------------------------------------------------------    
  ! Here we can compare the internal code parameters to the real physical ones 
  subroutine matchchi2(par,chi2pass)
    use constants
    use chi2calc
    use profile

    double precision :: par(:)
    double precision :: chi2pass, conc, rs, dc, m200

    m200 = par(1)
    conc = par(2)
    rs = rsnfw(m200,conc)
    dc = deltac(conc)
    chi2pass = chi2(rs,dc)
  
  end subroutine match2chi

  !--------------------------------------------------------------------------    
  ! This routine checks if MCMC candidate is within the given parameter space
  subroutine check(par)
    double precision :: par(:)

    do i=1, npar
      if (par(i) .lt. lobound(i)) then
        par(i) = lobound(i)
      elseif (par(i) .gt. upbound(i)) then
        par(i) = upbound(i)
      endif
    enddo
  
  end subroutine check
end module mcmc
!###########################################################################################################################
